/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "cff.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "charstring.h"
#include "../zaaklad/typy.h"

#define k32(u) ((((u) >> 24) & 0x000000ff) \
                | (((u) << 8) & 0x00ff0000) \
                | (((u) >> 8) & 0x0000ff00) \
                | (((u) << 24) & 0xff000000))

#define k16(u) ((((u) << 8) & 0xff00) \
                | (((u) >> 8) & 0x00ff))

#define retuj(n) ret = n; if (ret) return (ret)

Card8 cshCard8(FILE *f)
{
    Card8 b;

    if (fread(&b, sizeof(Card8), 1, f) != 1)
    {
        printf("Chyba pri cshiitanii!\n");
    }

    return (b);
}

Card16 cshCard16(FILE *f)
{
    Card16 b;

    if (fread(&b, sizeof(Card16), 1, f) != 1)
    {
        printf("Chyba pri cshiitanii!\n");
    }

    b = k16(b);

    return (b);
}

Offset cshOffset(FILE *f, OffSize offSize)
{
    Offset b = 0;

    if (fread(&b, offSize, 1, f) != 1)
    {
        printf("Chyba pri cshiitanii!\n");
    }

    switch (offSize) {
        case 1: return (b);
        case 2: return (k16(b));
        case 3: return (k32(b) >> 8);
        case 4: return (k32(b));
        default: {
            printf("Neznaamy offSize: %u\n", offSize);
            return (0);
        }
    }
}

OffSize cshOffSize(FILE *f)
{
    OffSize b;

    if (fread(&b, sizeof(OffSize), 1, f) != 1)
    {
        printf("Chyba pri cshiitanii!\n");
    }

    return (b);
}

SID cshSID(FILE *f)
{
    SID b;

    if (fread(&b, sizeof(SID), 1, f) != 1)
    {
        printf("Chyba pri cshiitanii!\n");
    }

    b = k16(b);

    return (b);
}

void print_header(Header *header)
{
    printf("major: %u\n", header->major);
    printf("minor: %u\n", header->minor);
    printf("hdrSize: %u\n", header->hdrSize);
    printf("offsize: %u\n", header->offsize);
}

void print_index(INDEX *index)
{
    unsigned long i;

    printf("index->count: %u\n", index->count);
    if (index->count == 0) return;

    printf("index->offsize: %u\n", index->offSize);
    printf("index->offset: ");
    for (i = 0; i < index->count + 1; i++)
    {
        printf("%lu", index->offset[i]);
        if (i + 1 < index->count + 1)
        {
            printf(", ");
        }
    }
    printf("\n");

    printf("index->data: ");
    for (i = 0; i < index->offset[index->count] - (index->offset[0]); i++)
    {
        printf("%c", index->data[i]);
    }
    printf("\n");
}

void print_index2(INDEX *index)
{
    unsigned long i, j;

    printf("index->count: %u\n", index->count);
    if (index->count == 0) return;

    printf("index->offsize: %u\n", index->offSize);
    printf("index->offset: ");
    for (i = 0; i < index->count + 1; i++)
    {
        printf("%lu", index->offset[i]);
        if (i + 1 < index->count + 1)
        {
            printf(", ");
        }
    }
    printf("\n");

    for (i = 0; i < index->count; i++)
    {
        printf("index->daata[%lu]: ", i);
        for (j = 0; j < index->velkosti[i]; j++)
        {
            printf("%c", index->daata[i][j]);
        }
        printf("\n");
    }
}

void print_dict(DICT *dict)
{
    DICT_K *dk = dict->z_dk;
    unsigned long i;

    while (dk)
    {
        if (dk->operator > 0xff)
        {
            printf(" operator: %u %u ",
                   dk->operator >> 8, dk->operator & 0xff);
        }
        else
        {
            printf(" operator: %u ", dk->operator);
        }
        printf("velkosth: %lu\n", dk->velkosth);
        printf("  daata: 0x");
        for (i = 0; i < dk->velkosth; i++)
        {
            printf("%02x", dk->daata[i]);
        }
        printf("\n");
        dk = dk->dk;
    }
}

void print_charset(Charset *charset)
{
    printf("charset->format: %u\n", charset->format);
    switch (charset->format) {
        case 0: {
            int i;

            printf("charset->daata.glyph: ");
            for (i = 0; i < charset->n; i++)
            {
                printf("%u", charset->daata.glyph[i]);
                if (i + 1 < charset->n)
                {
                    printf(", ");
                }
            }
            printf("\n");
        }
    }
}

int cshiitaj_header(FONT *font, Header *header)
{
    header->major = cshCard8(font->f);
    if (header->major > CFF_HEADER_MAJOR_VERSION)
    {
        printf("Neznaama verzia: %u\n", header->major);
        return (1);
    }
    header->minor = cshCard8(font->f);
    header->hdrSize = cshCard8(font->f);
    header->offsize = cshOffSize(font->f);

    return (0);
}

int spracuj_index(FONT *font, INDEX *index)
{
    int i;
    Offset vel_daata = index->offset[index->count] - (index->offset[0]);

    index->velkosti = calloc(index->count, sizeof(index->velkosti[0]));
    index->daata = calloc(index->count, sizeof(index->daata[0]));

    if ((index->velkosti == NULL) || (index->daata == NULL))
    {
        printf("Nepodarilo sa alokovath pamaeth!\n");
        return (1);
    }

    for (i = 0; i < index->count; i++)
    {
        index->velkosti[i] = index->offset[i + 1] - (index->offset[i]);
        index->daata[i] = malloc(sizeof(index->daata[i][0])
                                 * (index->velkosti[i]));
        if (index->daata[i] == NULL)
        {
            printf("Nepodarilo sa alokovath pamaeth!\n");
            return (1);
        }
        if ((index->offset[i] - 1 >= vel_daata)
            || (index->offset[i] + (index->velkosti[i]) - 1 > vel_daata))
        {
            printf("Niecsho nie je v poriadku s INDEXom!\n");
            return (1);
        }
        memcpy(index->daata[i],
               index->data + (index->offset[i] - 1),
               index->velkosti[i]);
    }

    return (0);
}

int cshiitaj_index(FONT *font, INDEX *index)
{
    int ret;
    int i;

    index->count = cshCard16(font->f);
    if (index->count == 0) return (0);

    index->offSize = cshOffSize(font->f);

    index->offset = calloc(index->count + 1, sizeof(index->offset[0]));
    if (index->offset == NULL)
    {
        printf("Nepodarilo sa alokovath pamaeth!\n");
        return (1);
    }

    for (i = 0; i < index->count + 1; i++)
    {
        index->offset[i] = cshOffset(font->f, index->offSize);
    }

    index->data = malloc(index->offset[index->count] - (index->offset[0]));
    if (index->data == NULL)
    {
        printf("Nepodarilo sa alokovath pamaeth!\n");
        return (1);
    }

    if (fread(index->data,
              sizeof(Card8),
              index->offset[index->count] - (index->offset[0]),
              font->f) != (index->offset[index->count] - (index->offset[0])))
    {
        printf("Chyba pri cshiitanii!\n");
        return (1);
    }

    retuj(spracuj_index(font, index));

    return (0);
}

int je_operand(Card8 bajt)
{
    if (bajt >= 0 && bajt <= 21) return (0);

    if ((bajt >= 22 && bajt <= 27)
        || (bajt == 31)
        || (bajt == 255))
    {
        printf("Rezervovanaa hodnota??? %u\n", bajt);
        return (2);
    }

    return (1);
}

int operandDllzhka(Card8 *daata, unsigned long velkosth)
{
    int i;

    if (velkosth == 0) return (0);

    if (daata[0] >= 32 && daata[0] <= 246) return (1);
    if (daata[0] >= 247 && daata[0] <= 250) return (2);
    if (daata[0] >= 251 && daata[0] <= 254) return (2);
    if (daata[0] == 28) return (3);
    if (daata[0] == 29) return (5);
    /* daata[0] == 30 */

    velkosth--;
    for (i = 1; velkosth; i++, velkosth--)
    {
        if ((daata[i] & 0xf) == 0xf)
        {
            i++;
            break;
        }
    }

    return (i);
}

long operandInteger(Card8 *daata, unsigned long velkosth)
{
    if (velkosth == 0) return (0);

    if (daata[0] >= 32 && daata[0] <= 246)
    {
        return (daata[0] - 139);
    }

    if (velkosth < 2)
    {
        printf("chyba operandInteger\n");
        return (0);
    }
    if (daata[0] >= 247 && daata[0] <= 250)
    {
        return ((daata[0] - 247) * 256 + daata[1] + 108);
    }
    if (daata[0] >= 251 && daata[0] <= 254)
    {
        return (-(daata[0] - 251) * 256 - daata[1] - 108);
    }

    if (velkosth < 3)
    {
        printf("chyba operandInteger\n");
        return (0);
    }
    if (daata[0] == 28)
    {
        return ((((long)(daata[1])) << 8) | ((long)(daata[2])));
    }

    if (velkosth < 5)
    {
        printf("chyba operandInteger\n");
        return (0);
    }
    if (daata[0] == 29)
    {
        return ((((long)(daata[1])) << 24) | (((long)(daata[2])) << 16)
                | (((long)(daata[3])) << 8) | ((long)(daata[4])));
    }

    printf("operand nie je integer\n");
    return (0);
}

void pridaj_dk(DICT *dict, DICT_K *dk)
{
    if (dict->z_dk == NULL)
    {
        dict->z_dk = dk;
        dict->k_dk = dk;
    }
    else
    {
        dict->k_dk->dk = dk;
        dict->k_dk = dk;
    }
    dk->dk = NULL;
}

DICT_K *dict_najdi(DICT *dict, Card16 operator)
{
    DICT_K *dk = dict->z_dk;

    while (dk)
    {
        if (dk->operator == operator) break;
        dk = dk->dk;
    }

    return (dk);
}

int dajDICT(DICT *dict, Card8 *daata, unsigned long velkosth)
{
    DICT_K *dk;
    unsigned long dk_velkosth;

    if (velkosth == 0) return (0);

    dk_velkosth = 0;
    while (velkosth)
    {
        if (je_operand(daata[0]))
        {
            int dllzhka = operandDllzhka(daata, velkosth);
            velkosth -= dllzhka;
            daata += dllzhka;
            dk_velkosth += dllzhka;
            continue;
        }

        dk = calloc(1, sizeof(*dk));
        if (dk == NULL)
        {
            printf("Nepodarilo sa alokovath pamaeth!\n");
            return (1);
        }

        dk->velkosth = dk_velkosth;
        dk->daata = malloc(dk->velkosth);
        if (dk->daata == NULL)
        {
            printf("Nepodarilo sa alokovath pamaeth!\n");
            return (1);
        }

        memcpy(dk->daata, daata - (dk->velkosth), dk->velkosth);

        if (daata[0] == 12)
        {
            if (velkosth <= 2)
            {
                printf("Chyba 12\n");
                free(dk);
                return (1);
            }
            dk->operator = (daata[0] << 8) | (daata[1]);
            velkosth--;
            daata++;
        }
        else
        {
            dk->operator = daata[0];
        };

        pridaj_dk(dict, dk);

        velkosth--;
        daata++;
        dk_velkosth = 0;
    }

    return (0);
}

int cshiitaj_charset(FONT *font, Charset *charset, Card16 nGlyphs)
{
    charset->format = cshCard8(font->f);
    switch (charset->format) {
        case 0: {
            int i;

            charset->n = nGlyphs - 1;
            charset->daata.glyph = calloc(charset->n,
                                          sizeof(charset->daata.glyph[0]));
            if (charset->daata.glyph == NULL)
            {
                printf("Nepodarilo sa alokovath pamaeth!\n");
                return (1);
            }
            for (i = 0; i < charset->n; i++)
            {
                charset->daata.glyph[i] = cshSID(font->f);
            }

            return (0);
        }
        default: {
            printf("Neznaamy format: %u\n", charset->format);
            return (1);
        }
    }
}

int priprav_na_charstringy(CFF *cff)
{
    DICT_K *dk;

    dk = dict_najdi(&(cff->Private), Private_DICT_defaultWidthX);
    if (dk == NULL) cff->defaultWidthX = 0;
    else cff->defaultWidthX = operandInteger(dk->daata, dk->velkosth);

    dk = dict_najdi(&(cff->Private), Private_DICT_nominalWidthX);
    if (dk == NULL) cff->nominalWidthX = 0;
    else cff->nominalWidthX = operandInteger(dk->daata, dk->velkosth);

    printf("defaultWidthX: %li nominalWidthX: %li\n",
           cff->defaultWidthX, cff->nominalWidthX);

    return (0);
}

int cffAnalyzuj(FONT *font, CFF_TAB *cff_tab)
{
    CFF *cff;
    int ret;

    cff = cff_tab->daata = calloc(1, sizeof(CFF));
    if (cff == NULL)
    {
        printf("Nepodarilo sa alokovath pamaeth!\n");
        return (1);
    }

    cff->f_shtart = ftell(font->f);
    if (cff->f_shtart == -1L)
    {
        printf("Chyba pri ftelle!\n");
        return (1);
    }

    retuj(cshiitaj_header(font, &(cff->header)));
    retuj(cshiitaj_index(font, &(cff->Name)));
    retuj(cshiitaj_index(font, &(cff->Top_DICT)));
    retuj(cshiitaj_index(font, &(cff->String)));
    retuj(cshiitaj_index(font, &(cff->Global_Subr)));

    dajDICT(&(cff->Top), cff->Top_DICT.daata[0], cff->Top_DICT.velkosti[0]);

    {
        DICT_K *dk = dict_najdi(&(cff->Top), Top_DICT_Private);
        Card8 *daata;
        unsigned long velkosth, zmena;
        long private_velkosth;
        long private_offset;

        if (dk == NULL)
        {
            printf("Chyyba Private\n");
            return (1);
        }
        daata = dk->daata;
        velkosth = dk->velkosth;

        private_velkosth = operandInteger(daata, velkosth);
        zmena = operandDllzhka(daata, velkosth);
        daata += zmena;
        velkosth -= zmena;
        private_offset = operandInteger(daata, velkosth);

        if (private_velkosth != 0)
        {
            daata = malloc(private_velkosth);
            if (daata == NULL)
            {
                printf("Nepodarilo sa alokovath pamaeth!\n");
                return (1);
            }

            if (fseek(font->f, cff->f_shtart + private_offset, SEEK_SET)
                != 0)
            {
                printf("Chyba pri fseeeku!\n");
                return (1);
            }

            if (fread(daata, sizeof(Card8), private_velkosth, font->f)
                != private_velkosth)
            {
                printf("Chyba pri cshiitanii!\n");
                return (1);
            }

            dajDICT(&(cff->Private), daata, private_velkosth);

            free(daata);

            dk = dict_najdi(&(cff->Private), Private_DICT_Subrs);
            if (dk)
            {
                long subrs_offset = operandInteger(dk->daata, dk->velkosth);

                if (fseek(font->f,
                          cff->f_shtart + private_offset + subrs_offset,
                          SEEK_SET)
                    != 0)
                {
                    printf("Chyba pri fseeeku!\n");
                    return (1);
                }

                retuj(cshiitaj_index(font, &(cff->Subr)));
            }
        }
    }

    {
        DICT_K *dk = dict_najdi(&(cff->Top), Top_DICT_CharStrings);
        long CharStrings_offset;

        if (dk == NULL)
        {
            printf("Chyyba CharStrings\n");
            return (1);
        }

        CharStrings_offset = operandInteger(dk->daata, dk->velkosth);

        if (fseek(font->f, cff->f_shtart + CharStrings_offset, SEEK_SET)
            != 0)
        {
            printf("Chyba pri fseeeku!\n");
            return (1);
        }

        retuj(cshiitaj_index(font, &(cff->CharStrings)));
    }

    {
        DICT_K *dk = dict_najdi(&(cff->Top), Top_DICT_charset);
        long charset_offset;

        if (dk == NULL)
        {
            printf("Chyyba charset\n");
            return (1);
        }

        charset_offset = operandInteger(dk->daata, dk->velkosth);

        if (fseek(font->f, cff->f_shtart + charset_offset, SEEK_SET)
            != 0)
        {
            printf("Chyba pri fseeeku!\n");
            return (1);
        }

        retuj(cshiitaj_charset(font, &(cff->charset), cff->CharStrings.count));
    }

    retuj(priprav_na_charstringy(cff));

    return (0);
}

void cffUvolhniIndex(INDEX *index)
{
    if (index->offset) free(index->offset);
    if (index->data) free(index->data);

    if (index->velkosti) free(index->velkosti);
    if (index->daata)
    {
        int i;

        for (i = 0; i < index->count; i++)
        {
            if (index->daata[i]) free(index->daata[i]);
        }

        free(index->daata);
    }
}

void cffUvolhniDict(DICT *dict)
{
    DICT_K *dk;

    while (dict->z_dk)
    {
        dk = dict->z_dk;
        dict->z_dk = dict->z_dk->dk;

        free(dk);
    }
}

void cffUvolhniCharset(Charset *charset)
{
    switch (charset->format) {
        case 0: if (charset->daata.glyph) free(charset->daata.glyph);
    }
}

void cffUvolhniCff(CFF_TAB *cff_tab)
{
    if (cff_tab->daata)
    {
        CFF *cff = cff_tab->daata;

        cffUvolhniIndex(&(cff->Name));
        cffUvolhniIndex(&(cff->Top_DICT));
        cffUvolhniIndex(&(cff->String));
        cffUvolhniIndex(&(cff->Global_Subr));
        cffUvolhniDict(&(cff->Top));
        cffUvolhniDict(&(cff->Private));
        cffUvolhniIndex(&(cff->Subr));
        cffUvolhniIndex(&(cff->CharStrings));
        cffUvolhniCharset(&(cff->charset));

        free(cff_tab->daata);
    }
}
