/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "charstring.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cff.h"

#define SKONTROLUJ_SMERNIIK_1(s) \
    do {if ((s) == NULL) { printf("Zlyy smerniik\n"); return (1); }} while(0)

#define retuj(n) ret = n; if (ret) return (ret)

typedef struct {
    CHSK obsah[LIMIT_Argument_stack];
    int vrch;
} AKUM;

void akum_inic(AKUM *akum)
{
    memset(akum->obsah, 0, sizeof(akum->obsah));
    akum->vrch = -1;
}

int akum_pridaj(AKUM *akum, CHSK *prvok)
{
    SKONTROLUJ_SMERNIIK_1(prvok);
    if (akum->vrch + 1 == LIMIT_Argument_stack)
    {
        printf("Priilish velha prvkov na akume\n");
        return (1);
    }

    akum->vrch++;
    akum->obsah[akum->vrch] = *prvok;

    return (0);
}

int identifikuj(Card8 *d, size_t velkosth)
{
    if (velkosth == 0)
    {
        printf("Chyba, velkosth == 0\n");
        return (-1);
    }
    if (d[0] == ESCAPE)
    {
        if (velkosth < 2)
        {
            printf("Chyba, velkosth < 2\n");
            return (-1);
        }
        return ((((int)d[0]) << 8) | d[1]);
    }

    if (((d[0] >= 32) && (d[0] <= 246))
        || ((d[0] >= 247) && (d[0] <= 250))
        || ((d[0] >= 251) && (d[0] <= 254)))
    {
        return (CHSK_INT);
    }

    return (d[0]);
}

void print_meno_op(int typ)
{
    switch(typ) {
        case OP_hstem: printf("hstem"); break;
        case OP_vstem: printf("vstem"); break;
        case OP_vmoveto: printf("vmoveto"); break;
        case OP_rlineto: printf("rlineto"); break;
        case OP_hlineto: printf("hlineto"); break;
        case OP_vlineto: printf("vlineto"); break;
        case OP_rrcurveto: printf("rrcurveto"); break;
        case OP_callsubr: printf("callsubr"); break;
        case OP_return: printf("return"); break;
        case OP_endchar: printf("endchar"); break;
        case OP_hstemhm: printf("hstemhm"); break;
        case OP_hintmask: printf("hintmask"); break;
        case OP_cntrmask: printf("cntrmask"); break;
        case OP_rmoveto: printf("rmoveto"); break;
        case OP_hmoveto: printf("hmoveto"); break;
        case OP_vstemhm: printf("vstemhm"); break;
        case OP_rcurveline: printf("rcurveline"); break;
        case OP_rlinecurve: printf("rlinecurve"); break;
        case OP_vvcurveto: printf("vvcurveto"); break;
        case OP_hhcurveto: printf("hhcurveto"); break;
        case OP_callgsubr: printf("callgsubr"); break;
        case OP_vhcurveto: printf("vhcurveto"); break;
        case OP_hvcurveto: printf("hvcurveto"); break;

        case OP_and: printf("and"); break;
        case OP_or: printf("or"); break;
        case OP_not: printf("not"); break;
        case OP_abs: printf("abs"); break;
        case OP_add: printf("add"); break;
        case OP_sub: printf("sub"); break;
        case OP_div: printf("div"); break;
        case OP_neg: printf("neg"); break;
        case OP_eq: printf("eq"); break;
        case OP_drop: printf("drop"); break;
        case OP_put: printf("put"); break;
        case OP_get: printf("get"); break;
        case OP_ifelse: printf("ifelse"); break;
        case OP_random: printf("random"); break;
        case OP_mul: printf("mul"); break;
        case OP_sqrt: printf("sqrt"); break;
        case OP_dup: printf("dup"); break;

        case OP_exch: printf("exch"); break;
        case OP_index: printf("index"); break;
        case OP_roll: printf("roll"); break;
        case OP_hflex: printf("hflex"); break;
        case OP_flex: printf("flex"); break;
        case OP_hflex1: printf("hflex1"); break;
        case OP_flex1: printf("flex1"); break;

        default: printf("Neznaama operaacia: %i", typ); break;
    }
}

void print_chsk(CHSK *chsk)
{
    if (chsk->typ > 0xff)
    {
        printf("typ: %i %i", chsk->typ >> 8, chsk->typ & 0xff);
        return;
    }
    printf("typ: %i", chsk->typ);
    switch (chsk->typ) {
        case CHSK_INT: {
            printf(" hodnota: %li\n", chsk->hodnota);
            return;
        }
        case CHSK_FIXED: {
            printf(" FIXED!!!!!\n");
            return;
        }
    }
    printf("\n");
}

int cshiitaj(CHSK *chsk, Card8 **sd, size_t *svelkosth)
{
    if ((*svelkosth) == 0)
    {
        printf("Chyba, (*svelkosth) == 0\n");
        return (1);
    }

    chsk->typ = identifikuj(*sd, *svelkosth);
    if (chsk->typ == -1) return (1);

    switch (chsk->typ) {
        case CHSK_INT: {
            if (((*sd)[0] >= 32) && ((*sd)[0] <= 246))
            {
                chsk->hodnota = (*sd)[0] - 139;
                (*sd)++;
                (*svelkosth)--;
                return (0);
            }

            if ((*svelkosth) < 2)
            {
                printf("(*svelkosth) < 2\n");
                return (0);
            }

            if (((*sd)[0] >= 247) && ((*sd)[0] <= 250))
            {
                chsk->hodnota = ((*sd)[0] - 247) * 256 + (*sd)[1] + 108;
                (*sd) += 2;
                (*svelkosth) -= 2;
                return (0);
            }
            if (((*sd)[0] >= 251) && ((*sd)[0] <= 254))
            {
                chsk->hodnota = -((*sd)[0] - 251) * 256 - (*sd)[1] - 108;
                (*sd) += 2;
                (*svelkosth) -= 2;
                return (0);
            }

            if ((*svelkosth) < 3)
            {
                printf("(*svelkosth) < 3\n");
                return (0);
            }

            chsk->hodnota = (((long)((*sd)[0])) << 8) | (*sd)[1];
            (*sd) += 3;
            (*svelkosth) -= 3;
            return (0);
        }
    }

    if (chsk->typ > 0xff)
    {
        (*sd) += 2;
        (*svelkosth) -= 2;
        return (0);
    }

    (*sd)++;
    (*svelkosth)--;
    return (0);
}

int vykonaj_hstem(CFF *cff, AKUM *akum)
{
    int i;

    if (akum->vrch < 1)
    {
        printf("(hstem) chyyba prvyy paar\n");
        return (1);
    }

    if (akum->vrch % 2 == 0)
    {
        printf("(hstem) nepaarny pocshet argumentov\n");
        return (1);
    }

    for (i = 0; i < (akum->vrch + 1) / 2; i++)
    {
        printf(" %li %li\n",
               akum->obsah[i].hodnota,
               akum->obsah[i + 1].hodnota);
    }
    akum->vrch = -1;

    return (0);
}

int vykonaj_vstem(CFF *cff, AKUM *akum)
{
    int i;

    if (akum->vrch < 1)
    {
        printf("(vstem) chyyba prvyy paar\n");
        return (1);
    }

    if (akum->vrch % 2 == 0)
    {
        printf("(vstem) nepaarny pocshet argumentov\n");
        return (1);
    }

    for (i = 0; i < (akum->vrch + 1) / 2; i++)
    {
        printf(" %li %li\n",
               akum->obsah[i].hodnota,
               akum->obsah[i + 1].hodnota);
    }
    akum->vrch = -1;

    return (0);
}

int vykonaj_hmoveto(CFF *cff, AKUM *akum)
{
    if (akum->vrch < 0)
    {
        printf("(hmoveto) chyyba argument\n");
        return (1);
    }

    cff->bodX += akum->obsah[0].hodnota;
    akum->vrch = -1;

    return (0);
}

int vykonaj_rmoveto(CFF *cff, AKUM *akum)
{
    if (akum->vrch < 1)
    {
        printf("(rmoveto) chyybajuu argumenty\n");
        return (1);
    }

    cff->bodX += akum->obsah[0].hodnota;
    cff->bodY += akum->obsah[1].hodnota;
    akum->vrch = -1;

    return (0);
}

void dajPixel(unsigned long *obraz, long x, long y)
{
    obraz[x + y * 640] = 0x00000000;
}

void hcshiara(CFF *cff, long x)
{
    long i;

    if (x < 0)
    {
        for (i = x; i; i++)
        {
            dajPixel(cff->obraz, cff->bodX + i, cff->bodY);
        }
    }
    else
    {
        for (i = x; i; i--)
        {
            dajPixel(cff->obraz, cff->bodX + i, cff->bodY);
        }
    }

    cff->bodX += x;
}

void vcshiara(CFF *cff, long y)
{
    long i;

    if (y < 0)
    {
        for (i = y; i; i++)
        {
            dajPixel(cff->obraz, cff->bodX, cff->bodY + i);
        }
    }
    else
    {
        for (i = y; i; i--)
        {
            dajPixel(cff->obraz, cff->bodX, cff->bodY + i);
        }
    }

    cff->bodY += y;
}

int vykonaj_hlineto(CFF *cff, AKUM *akum)
{
    int i;

    if (akum->vrch < 0)
    {
        printf("(hlineto) chyyba argument\n");
        return (1);
    }

    for (i = 0; i < akum->vrch + 1; i++)
    {
        if (i % 2 == 0)
        {
            hcshiara(cff, akum->obsah[i].hodnota);
        }
        else
        {
            vcshiara(cff, akum->obsah[i].hodnota);
        }
    }
    akum->vrch = -1;

    return (0);
}

int vykonaj_vlineto(CFF *cff, AKUM *akum)
{
    int i;

    if (akum->vrch < 0)
    {
        printf("(vlineto) chyyba argument\n");
        return (1);
    }

    for (i = 0; i < akum->vrch + 1; i++)
    {
        if (i % 2 == 0)
        {
            vcshiara(cff, akum->obsah[i].hodnota);
        }
        else
        {
            hcshiara(cff, akum->obsah[i].hodnota);
        }
    }
    akum->vrch = -1;

    return (0);
}

int vykonaj(CFF *cff, AKUM *akum, int typ)
{
    printf("op: ");
    print_meno_op(typ);
    printf(" %i\n", typ);

    switch (typ) {
        case OP_hstem: return(vykonaj_hstem(cff, akum));
        case OP_vstem: return(vykonaj_vstem(cff, akum));
        case OP_hmoveto: return(vykonaj_hmoveto(cff, akum));
        case OP_rmoveto: return(vykonaj_rmoveto(cff, akum));
        case OP_hlineto: return(vykonaj_hlineto(cff, akum));
        case OP_vlineto: return(vykonaj_vlineto(cff, akum));
    }

    return (0);
}

int pracuj(CFF *cff, unsigned long *obraz)
{
    INDEX *CharStrings = &(cff->CharStrings);
    int glyf = 0;
    long width;
    int i;
    Card8 *daata = CharStrings->daata[glyf];
    size_t velkosth = CharStrings->velkosti[glyf];
    CHSK chsk;
    AKUM akum;
    int ret;

    cff->bodX = 0;
    cff->bodY = 50;
    cff->obraz = obraz;

    akum_inic(&akum);

    printf("CharStrings->daata[%i]: 0x", glyf);
    for (i = 0; i < CharStrings->velkosti[glyf]; i++)
    {
        printf("%02x", CharStrings->daata[glyf][i]);
    }
    printf("\n");

    cshiitaj(&chsk, &daata, &velkosth);
    width = cff->nominalWidthX + (chsk.hodnota);
    printf("width: %li\n", width);

    while (velkosth)
    {
        cshiitaj(&chsk, &daata, &velkosth);
        print_chsk(&chsk);

        switch (chsk.typ) {
            case CHSK_INT:
            case CHSK_FIXED: {
                retuj(akum_pridaj(&akum, &chsk));
                break;
            }

            case OP_endchar: {
                return (0);
            }

            default: {
                retuj(vykonaj(cff, &akum, chsk.typ));
                break;
            }
        }
    }

    return (0);
}
