/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "taby.h"

#include <stdio.h>
#include <stdlib.h>

#include "../zaaklad/typy.h"
#include "../cff/cff.h"

#define k32(u) ((((u) >> 24) & 0x000000ff) \
                | (((u) << 8) & 0x00ff0000) \
                | (((u) >> 8) & 0x0000ff00) \
                | (((u) << 24) & 0xff000000))

static USHORT ushortMax(USHORT a, USHORT b)
{
    return (a > b ? a : b);
}

void printTag(ULONG tag)
{
    int i;
    char *p = (char *)&tag;

    tag = k32(tag);

    for (i = 0; i < 4; i++)
    {
        printf("%c", p[i]);
    }
}

void printFixed(Fixed fixed)
{
    printf("%i.%05i",
           (int)((fixed >> 16) & 0xffff),
           (int)(fixed & 0xffff));
}

void printLongdatetime(LONGDATETIME *ldt)
{
    int i;

    for (i = 0; i < 8; i++)
    {
        printf("%02x", ldt->daata[i]);
    }
    printf(" +++mozhno zle");
}

void print_font(FONT *font)
{
    int i;

    printf("sfnt_version: ");
    printTag(font->sfnt_version);
    printf(" 0x%08lx\n", font->sfnt_version);
    printf("numTables: %u\n", font->numTables);
    printf("searchRange: %u\n", font->searchRange);
    printf("entrySelector: %u\n", font->entrySelector);
    printf("rangeShift: %u\n", font->rangeShift);

    for (i = 0; i < font->numTables; i++)
    {
        printf(" ");
        printTag(font->taby[i].tag);
        printf(" 0x%08lx ", font->taby[i].tag);
        printf(" checkSum: %lu", font->taby[i].checkSum);
        printf(" Offset: %lu", font->taby[i].Offset);
        printf(" length: %lu", font->taby[i].length);
        printf("\n");
    }
}

void print_cmap(CMAP_TAB *cmap)
{
    int i;

    printf("Version: %u\n", cmap->Version);
    printf("numTables: %u\n", cmap->numTables);

    if (cmap->cmap_subtaby == NULL)
    {
        printf("cmap->cmap_subtaby je NULL\n");
        return;
    }

    for (i = 0; i < cmap->numTables; i++)
    {
        CMAP_SUBTAB *cs = cmap->cmap_subtaby + i;

        printf("Subtable %i:\n", i);
        printf(" platformID: %u\n", cs->platformID);
        printf(" encodingID: %u\n", cs->encodingID);
        printf(" Offset: %lu\n", cs->Offset);
        printf("\n");

        printf(" format: %u\n", cs->format);
        printf(" length: %u\n", cs->length);
        printf(" language: %u\n", cs->language);
        printf("\n");
    }
}

void print_head(HEAD_TAB *head)
{
    printf("Table_version_number: 0x%08lx\n", head->Table_version_number);
    printf("fontRevision: 0x%08lx %lu\n", head->fontRevision,
           head->fontRevision);
    printf("checkSumAdjustment: 0x%08lx %lu\n", head->checkSumAdjustment,
           head->checkSumAdjustment);
    printf("magicNumber: 0x%08lx\n", head->magicNumber);
    printf("flags: 0x%04x\n", head->flags);
    printf("unitsPerEm: %u\n", head->unitsPerEm);

    printf("created: ");
    printLongdatetime(&(head->created));
    printf("\n");
    printf("modified: ");
    printLongdatetime(&(head->modified));
    printf("\n");

    printf("xMin: %i\n", head->xMin);
    printf("yMin: %i\n", head->yMin);
    printf("xMax: %i\n", head->xMax);
    printf("yMax: %i\n", head->yMax);

    printf("macStyle: %u\n", head->macStyle);
    printf("lowestRecPPEM: %u\n", head->lowestRecPPEM);

    printf("fontDirectionHint: %i\n", head->fontDirectionHint);
    printf("indexToLocFormat: %i\n", head->indexToLocFormat);
    printf("glyphDataFormat: %i\n", head->glyphDataFormat);
}

void print_hhea(HHEA_TAB *hhea)
{
    printf("Table_version_number: 0x%08lx\n", hhea->Table_version_number);
    printf("Ascender: %i\n", hhea->Ascender);
    printf("Descender: %i\n", hhea->Descender);
    printf("LineGap: %i\n", hhea->LineGap);
    printf("advanceWidthMax: %u\n", hhea->advanceWidthMax);

    printf("minLeftSideBearing: %i\n", hhea->minLeftSideBearing);
    printf("minRightSideBearing: %i\n", hhea->minRightSideBearing);
    printf("xMaxExtend: %i\n", hhea->xMaxExtend);

    printf("caretSlopeRise: %i\n", hhea->caretSlopeRise);
    printf("caretSlopeRun: %i\n", hhea->caretSlopeRun);
    printf("caretOffset: %i\n", hhea->caretOffset);

    printf("metricDataFormat: %i\n", hhea->metricDataFormat);
    printf("numberOfHMetrics: %u\n", hhea->numberOfHMetrics);
}

void print_hmtx(HMTX_TAB *hmtx, FONT *font)
{
    int i;
    HHEA_TAB *hhea;
    MAXP_TAB *maxp;

    hhea = daj_hhea(font);
    if (hhea == NULL)
    {
        printf("Chyyba hhea tab.!\n");
        return;
    }

    maxp = daj_maxp(font);
    if (maxp == NULL)
    {
        printf("Chyyba maxp tab.!\n");
        return;
    }

    for (i = 0; i < hhea->numberOfHMetrics; i++)
    {
        printf("longHorMetric %i\n", i);
        printf(" advanceWidth: %u\n", hmtx->hMetrics[i].advanceWidth);
        printf(" lsb: %i\n", hmtx->hMetrics[i].lsb);
    }

    for (i = 0; i < maxp->numGlyphs - (hhea->numberOfHMetrics); i++)
    {
        printf("leftSideBearing: %i\n", hmtx->leftSideBearing[i]);
    }
}

void print_maxp(MAXP_TAB *maxp)
{
    printf("Table_version_number: 0x%08lx\n", maxp->Table_version_number);
    printf("numGlyphs: %u\n", maxp->numGlyphs);
    if (maxp->Table_version_number == MAXP_TRUETYPE)
    {
        printf("maxPoints: %u\n", maxp->maxPoints);
        printf("maxContours: %u\n", maxp->maxContours);
        printf("maxCompositePoints: %u\n", maxp->maxCompositePoints);
        printf("maxCompositeContours: %u\n", maxp->maxCompositeContours);
        printf("maxZones: %u\n", maxp->maxZones);
        printf("maxTwilightPoints: %u\n", maxp->maxTwilightPoints);
        printf("maxStorage: %u\n", maxp->maxStorage);
        printf("maxFunctionDefs: %u\n", maxp->maxFunctionDefs);
        printf("maxInstructionDefs: %u\n", maxp->maxInstructionDefs);
        printf("maxStackElements: %u\n", maxp->maxStackElements);
        printf("maxSizeOfInstructions: %u\n", maxp->maxSizeOfInstructions);
        printf("maxComponentElements: %u\n", maxp->maxComponentElements);
        printf("maxComponentDepth: %u\n", maxp->maxComponentDepth);
    }
}

void print_os_2(OS_2_TAB *os_2)
{
    printf("Version: 0x%04x\n", os_2->Version);
    printf("xAvgCharWidth: %i\n", os_2->xAvgCharWidth);
    printf("usWeightClass: %u\n", os_2->usWeightClass);
    printf("usWidthClass: %u\n", os_2->usWidthClass);
    printf("fsType: %u\n", os_2->fsType);
    printf("ySubscriptXSize: %i\n", os_2->ySubscriptXSize);
    printf("ySubscriptYSize: %i\n", os_2->ySubscriptYSize);
    printf("ySubscriptXOffset: %i\n", os_2->ySubscriptXOffset);
    printf("ySubscriptYOffset: %i\n", os_2->ySubscriptYOffset);
    printf("ySuperscriptXSize: %i\n", os_2->ySuperscriptXSize);
    printf("ySuperscriptYSize: %i\n", os_2->ySuperscriptYSize);
    printf("ySuperscriptXOffset: %i\n", os_2->ySuperscriptXOffset);
    printf("ySuperscriptYOffset: %i\n", os_2->ySuperscriptYOffset);
    printf("yStrikeoutSize: %i\n", os_2->yStrikeoutSize);
    printf("yStrikeoutPosition: %i\n", os_2->yStrikeoutPosition);
    printf("sFamilyClass: %i\n", os_2->sFamilyClass);

    {
        int i;
        printf("Panose: 0x");
        for (i = 0; i < sizeof(os_2->Panose) / sizeof(os_2->Panose[0]); i++)
        {
            printf("%02x", os_2->Panose[i]);
        }
        printf("\n");
    }

    printf("ulUnicodeRange1: 0x%08lx\n", os_2->ulUnicodeRange1);
    if (os_2->Version >= 0x0001)
    {
        printf("ulUnicodeRange1: 0x%08lx\n", os_2->ulUnicodeRange2);
        printf("ulUnicodeRange2: 0x%08lx\n", os_2->ulUnicodeRange3);
        printf("ulUnicodeRange3: 0x%08lx\n", os_2->ulUnicodeRange4);
    }

    {
        int i;
        printf("achVendID: ");
        for (i = 0;
             i < sizeof(os_2->achVendID) / sizeof(os_2->achVendID[0]);
             i++)
        {
            printf("%c", os_2->achVendID[i]);
        }
        printf("\n");
    }

    printf("fsSelection: 0x%04x\n", os_2->fsSelection);
    printf("usFirstCharIndex: %u\n", os_2->usFirstCharIndex);
    printf("usLastCharIndex: %u\n", os_2->usLastCharIndex);
    printf("sTypoAscender: %i\n", os_2->sTypoAscender);
    printf("sTypoDescender: %i\n", os_2->sTypoDescender);
    printf("sTypoLineGap: %i\n", os_2->sTypoLineGap);
    printf("usWinAscent: %u\n", os_2->usWinAscent);
    printf("usWinDescent: %u\n", os_2->usWinDescent);

    if (os_2->Version >= 0x0001)
    {
        printf("ulCodePageRange1: 0x%08lx\n", os_2->ulCodePageRange1);
        printf("ulCodePageRange2: 0x%08lx\n", os_2->ulCodePageRange2);
    }

    if (os_2->Version >= 0x0002)
    {
        printf("sxHeight: %i\n", os_2->sxHeight);
        printf("sCapHeight: %i\n", os_2->sCapHeight);
        printf("usDefaultChar: %u\n", os_2->usDefaultChar);
        printf("usBreakChar: %u\n", os_2->usBreakChar);
        printf("usMaxContext: %u\n", os_2->usMaxContext);
    }
}

void print_post(POST_TAB *post)
{
    printf("Version: 0x%08lx\n", post->Version);
    printf("italicAngle: ");
    printFixed(post->italicAngle);
    printf("\n");
    printf("underlinePosition: %i\n", post->underlinePosition);
    printf("underlinePosition: %i\n", post->underlineThickness);
    printf("isFixedPitch: %lu\n", post->isFixedPitch);
    printf("minMemType42: %lu\n", post->minMemType42);
    printf("maxMemType42: %lu\n", post->maxMemType42);
    printf("minMemType1: %lu\n", post->minMemType1);
    printf("maxMemType1: %lu\n", post->maxMemType1);

    if (post->Version == 0x00020000)
    {
        int i;
        USHORT numberNewGlyphs = 0;

        printf("numberOfGlyphs: %u\n", post->numberOfGlyphs);
        for (i = 0; i < post->numberOfGlyphs; i++)
        {
            printf(" glyphNameIndex[%i]: %u\n", i, post->glyphNameIndex[i]);
            numberNewGlyphs = ushortMax(numberNewGlyphs,
                                        post->glyphNameIndex[i]);
        }
        /* Prvyych 258 glyfov je v shtandarnej Makintosh sade. */
        numberNewGlyphs = numberNewGlyphs - 258 + 1;

        for (i = 0; i < numberNewGlyphs; i++)
        {
            printf(" names[%i}: %s\n", i, post->names[i]);
        }
    }
}

void print_loca(LOCA_TAB *loca, FONT *font)
{
    MAXP_TAB *maxp;
    int i;

    maxp = daj_maxp(font);
    if (maxp == NULL)
    {
        printf("Chyyba maxp tab.!\n");
        return;
    }

    if (loca->Offsets)
    {
        for (i = 0; i < maxp->numGlyphs + 1; i++)
        {
            printf("loca->Offsets[%i]: %lu\n", i, loca->Offsets[i]);
        }
    }
}

void print_cvt(CVT_TAB *cvt)
{
    int i;

    if (cvt->hodnoty)
    {
        for (i = 0; i < cvt->n; i++)
        {
            printf("cvt->hodnoty[%i]: %i\n", i, cvt->hodnoty[i]);
        }
    }
}

void print_fpgm(FPGM_TAB *fpgm)
{
    int i;

    if (fpgm->Instructions)
    {
        for (i = 0; i < fpgm->n; i++)
        {
            printf("fpgm->Instructions[%i]: 0x%02x\n",
                   i, fpgm->Instructions[i]);
        }
    }
}

void print_prep(PREP_TAB *prep)
{
    int i;

    if (prep->Instructions)
    {
        for (i = 0; i < prep->n; i++)
        {
            printf("prep->Instructions[%i]: 0x%02x\n",
                   i, prep->Instructions[i]);
        }
    }
}

void print_simple_glyph(GLYPH *glyph)
{
    SIMPLE_GLYPH_DESCRIPTION *sgd = &(glyph->Description.sgd);
    int i;

    if (sgd->endPtsOfContours)
    {
        printf("  sgd->endPtsOfContours: ");
        for (i = 0; i < glyph->numberOfContours; i++)
        {
            printf("%u", sgd->endPtsOfContours[i]);
            if (i + 1 < glyph->numberOfContours)
            {
                printf(", ");
            }
        }
        printf("\n");
    }

    printf("  sgd->instructionLength: %u\n", sgd->instructionLength);

    if (sgd->endPtsOfContours)
    {
        printf("  sgd->instructions: ");
        for (i = 0; i < sgd->instructionLength; i++)
        {
            printf("0x%02x", sgd->instructions[i]);
            if (i + 1 < sgd->instructionLength)
            {
                printf(", ");
            }
        }
        printf("\n");
    }

    if (sgd->flags)
    {
        printf("  sgd->flags: ");
        for (i = 0; i < sgd->pocshetBodov; i++)
        {
            printf("0x%02x", sgd->flags[i]);
            if (i + 1 < sgd->pocshetBodov)
            {
                printf(", ");
            }
        }
        printf("\n");
    }


    if (sgd->xCoordinates)
    {
        printf("  sgd->xCoordinates: ");
        for (i = 0; i < sgd->pocshetBodov; i++)
        {
            printf("%i", sgd->xCoordinates[i]);
            if (i + 1 < sgd->pocshetBodov)
            {
                printf(", ");
            }
        }
        printf("\n");
    }

    if (sgd->yCoordinates)
    {
        printf("  sgd->yCoordinates: ");
        for (i = 0; i < sgd->pocshetBodov; i++)
        {
            printf("%i", sgd->yCoordinates[i]);
            if (i + 1 < sgd->pocshetBodov)
            {
                printf(", ");
            }
        }
        printf("\n");
    }
}

void print_glyph(GLYPH *glyph)
{
    printf(" numberOfContours: %i\n", glyph->numberOfContours);
    printf(" xMin: %i\n", glyph->xMin);
    printf(" yMin: %i\n", glyph->yMin);
    printf(" xMax: %i\n", glyph->xMax);
    printf(" yMax: %i\n", glyph->yMax);

    if (glyph->numberOfContours > 0)
    {
        print_simple_glyph(glyph);
    }
    else if (glyph->numberOfContours < 0)
    {
        /*printf("Compound\n");*/
    }
}

void print_glyf(GLYF_TAB *glyf)
{
    int i;

    if (glyf->glyphy)
    {
        for (i = 0; i < glyf->n; i++)
        {
            printf("glyphy[%i]:\n", i);
            print_glyph(glyf->glyphy + i);
        }
    }
}

#define najdi_tab(tab, tab_tag) \
{ \
    int i; \
\
    for (i = 0; i < font->numTables; i++) \
    {\
        if (font->taby[i].tag == tab_tag \
            && font->taby[i].stav == STAV_HOTOVEE) \
        { \
            return (&(font->taby[i].daata.tab)); \
        } \
    } \
\
    return (NULL); \
} \

CMAP_TAB *daj_cmap(FONT *font) najdi_tab(cmap, TAG_CMAP)
HEAD_TAB *daj_head(FONT *font) najdi_tab(head, TAG_HEAD)
HHEA_TAB *daj_hhea(FONT *font) najdi_tab(hhea, TAG_HHEA)
HMTX_TAB *daj_hmtx(FONT *font) najdi_tab(hmtx, TAG_HMTX)
MAXP_TAB *daj_maxp(FONT *font) najdi_tab(maxp, TAG_MAXP)
OS_2_TAB *daj_os_2(FONT *font) najdi_tab(os_2, TAG_OS_2)
POST_TAB *daj_post(FONT *font) najdi_tab(post, TAG_POST)
LOCA_TAB *daj_loca(FONT *font) najdi_tab(loca, TAG_LOCA)
CVT_TAB *daj_cvt(FONT *font) najdi_tab(cvt, TAG_CVT)
FPGM_TAB *daj_fpgm(FONT *font) najdi_tab(fpgm, TAG_FPGM)
PREP_TAB *daj_prep(FONT *font) najdi_tab(prep, TAG_PREP)
GLYF_TAB *daj_glyf(FONT *font) najdi_tab(glyf, TAG_GLYF)
CFF_TAB *daj_cff(FONT *font) najdi_tab(cff, TAG_CFF)

void cmapUvolhni(CMAP_TAB *cmap)
{
    int i;

    if (cmap->cmap_subtaby)
    {
        for (i = 0; i < cmap->numTables; i++)
        {
            CMAP_SUBTAB *cs = cmap->cmap_subtaby + i;

            if (cs->glyf_indexy)
            {
                free(cs->glyf_indexy);
            }
        }
        free(cmap->cmap_subtaby);
    }
}

void hmtxUvolhni(HMTX_TAB *hmtx)
{
    if (hmtx->hMetrics) free(hmtx->hMetrics);
    if (hmtx->leftSideBearing) free(hmtx->leftSideBearing);
}

void postUvolhni(POST_TAB *post)
{
    if (post->Version == 0x00020000)
    {
        if (post->glyphNameIndex)
        {
            if (post->names)
            {
                int i;
                USHORT numberNewGlyphs = 0;

                for (i = 0; i < post->numberOfGlyphs; i++)
                {
                    numberNewGlyphs = ushortMax(numberNewGlyphs,
                                                post->glyphNameIndex[i]);
                }
                /* Prvyych 258 glyfov je v shtandarnej Makintosh sade. */
                numberNewGlyphs = numberNewGlyphs - 258 + 1;
                for (i = 0; i < numberNewGlyphs; i++)
                {
                    if (post->names[i]) free(post->names[i]);
                }

                free(post->names);
            }
            free(post->glyphNameIndex);
        }
    }
}

void locaUvolhni(LOCA_TAB *loca)
{
    if (loca->Offsets) free(loca->Offsets);
}

void cvtUvolhni(CVT_TAB *cvt)
{
    if (cvt->hodnoty) free(cvt->hodnoty);
}

void fpgmUvolhni(FPGM_TAB *fpgm)
{
    if (fpgm->Instructions) free(fpgm->Instructions);
}

void prepUvolhni(PREP_TAB *prep)
{
    if (prep->Instructions) free(prep->Instructions);
}

void glyphUvolhni(GLYPH *glyph)
{
    if (glyph->numberOfContours > 0)
    {
        SIMPLE_GLYPH_DESCRIPTION *sgd = &(glyph->Description.sgd);

        if (sgd->endPtsOfContours) free(sgd->endPtsOfContours);
        if (sgd->instructions) free(sgd->instructions);
        if (sgd->flags) free(sgd->flags);
        if (sgd->xCoordinates) free(sgd->xCoordinates);
        if (sgd->yCoordinates) free(sgd->yCoordinates);
    }
}

void glyfUvolhni(GLYF_TAB *glyf)
{
    if (glyf->glyphy) free(glyf->glyphy);
}

void cffUvolhni(CFF_TAB *cff)
{
    cffUvolhniCff(cff);
}

void tabulhkaUvolhni(TABLE_RECORD *tab)
{
    switch (tab->tag) {
        case TAG_CMAP: {
            cmapUvolhni(&(tab->daata.cmap));

            break;
        }

        case TAG_HMTX: {
            hmtxUvolhni(&(tab->daata.hmtx));

            break;
        }

        case TAG_POST: {
            postUvolhni(&(tab->daata.post));

            break;
        }

        case TAG_LOCA: {
            locaUvolhni(&(tab->daata.loca));

            break;
        }

        case TAG_CVT: {
            cvtUvolhni(&(tab->daata.cvt));

            break;
        }

        case TAG_FPGM: {
            fpgmUvolhni(&(tab->daata.fpgm));

            break;
        }

        case TAG_PREP: {
            prepUvolhni(&(tab->daata.prep));

            break;
        }

        case TAG_GLYF: {
            glyfUvolhni(&(tab->daata.glyf));

            break;
        }

        case TAG_CFF: {
            cffUvolhni(&(tab->daata.cff));

            break;
        }

        case TAG_HEAD:
        case TAG_HHEA:
        case TAG_MAXP:
        case TAG_OS_2: {
            /* Netreba spravith nicsh. */
            break;
        }
    }
}
